import logging
from functools import reduce
from typing import Any, Dict, List, Optional, Tuple

import datayoga_core.blocks.redis.utils as redis_utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context
from datayoga_core.utils import get_connection_details

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.command = self.properties.get("command")
        self.key_field = self.properties.get("key_field")
        connection = get_connection_details(self.properties.get("connection"), context)
        self.redis_client = redis_utils.get_client(
            connection.get("host"),
            connection.get("port"),
            connection.get("password"))
        logger.info(f"Writing to Redis connection '{self.properties.get('connection')}'")

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        pipeline = self.redis_client.pipeline()
        for record in data:
            dict_as_list = list(reduce(lambda x, y: x + y, record.items()))
            pipeline.execute_command(self.command, record[self.key_field], *dict_as_list)
        pipeline.execute()
        # TODO: check the return value from the pipeline
        return data, [Result.SUCCESS]*len(data)



import json
import logging
import sys
from typing import Any, Dict, List, Optional, Tuple

from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Block(DyBlock):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        for record in data:
            # remove the internal $$msg_id column
            filtered_record = {i: record[i] for i in record if i != Block.MSG_ID_FIELD}
            sys.stdout.write(f"{json.dumps(filtered_record)}\n")

        # if we made it here, it is a success. return the data and the success result
        return data, [Result.SUCCESS]*len(data)



from typing import Optional

import redis
from redis import Redis


def get_client(host: str, port: int, password: Optional[str] = None) -> Redis:
    try:
        client = redis.Redis(
            host=host,
            port=port,
            password=password,
            decode_responses=True,
            client_name="datayoga"
        )

        client.ping()
        return client
    except Exception as e:
        raise ValueError(f"can not connect to Redis on {host}:{port}:\n {e}")



import logging
from typing import Any, Dict, List, Optional, Tuple

import sqlalchemy as sa
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context
from datayoga_core.utils import get_connection_details

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")

        connection = get_connection_details(self.properties.get("connection"), context)

        engine_url = sa.engine.URL.create(
            drivername=connection.get("type"),
            host=connection.get("host"),
            port=connection.get("port"),
            username=connection.get("user"),
            password=connection.get("password"),
            database=connection.get("database")
        )

        self.schema = self.properties.get("schema")
        self.table = self.properties.get("table")
        self.engine = sa.create_engine(engine_url, echo=False)
        logger.debug(f"Connecting to {connection.get('type')}")

        self.conn = self.engine.connect()

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        logger.debug(f"Running {self.get_block_name()}")

        logger.info(f"Inserting {len(data)} record(s) to {self.table} table")
        tbl = sa.Table(self.table, sa.MetaData(schema=self.schema), autoload_with=self.engine)

        self.conn.execute(tbl.insert(), data)
        # if we made it here, it is a success. return the data and the success result
        return data, [Result.SUCCESS]*len(data)



import logging
from typing import Any, Dict, List, Tuple

from datayoga_core import expression, utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Context = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

        self.fields = {}
        for property in self.properties["fields"]:
            self.fields[property["field"]] = expression.compile(
                property["language"],
                property["expression"])

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        logger.debug(f"Running {self.get_block_name()}")
        results = []
        for row in data:
            for field in self.fields:
                obj = row
                field_path = utils.split_field(field)

                for key in field_path[:-1]:
                    key = utils.unescape_field(key)
                    if key in obj:
                        obj = obj[key]
                    else:
                        obj[key] = {}

                obj[utils.unescape_field(field_path[-1:][0])] = self.fields[field].search(row)
            results.append(Result.SUCCESS)
        return data, results



import pytest
from datayoga_core.block import Result
from datayoga_core.blocks.add_field.block import Block


@pytest.mark.asyncio
async def test_add_field():
    block = Block({"field": "full_name",
                   "language": "jmespath",
                   "expression": "[fname,lname] | join(' ', @)"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == ([
        {"fname": "john", "lname": "doe", "full_name": "john doe"}], [Result.SUCCESS]
    )


@pytest.mark.asyncio
async def test_add_nested_field():
    block = Block({"field": "name.full_name",
                   "language": "jmespath",
                  "expression": "[name.fname,name.lname] | join(' ', @)"})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == ([
        {"name": {"fname": "john", "lname": "doe", "full_name": "john doe"}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_add_multiple_fields():
    block = Block({"fields": [{"field": "name.full_name", "language": "jmespath", "expression": "concat([name.fname, ' ', name.lname])"}, {
                  "field": "name.fname_upper", "language": "jmespath", "expression": "upper(name.fname)"}]})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == ([
        {"name": {"fname": "john", "lname": "doe", "full_name": "john doe", "fname_upper": "JOHN"}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_add_field_with_dot():
    block = Block({"field": "name\.full_name",
                   "language": "jmespath",
                   "expression": "[fname,lname] | join(' ', @)"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == ([
        {"fname": "john", "lname": "doe", "name.full_name": "john doe"}], [Result.SUCCESS])



import logging
from typing import Any, Dict, List, Optional

from datayoga_core.context import Context
from datayoga_core.job import Job, compile_job, validate_job

logger = logging.getLogger("dy")


def compile(
        job_settings: Dict[str, Any],
        whitelisted_blocks: Optional[List[str]] = None) -> Job:
    """
    Compiles a job in YAML 

    Args:
        job_settings (Dict[str, Any]): Job settings
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Returns:
        Job: Compiled job
    """
    logger.debug("Compiling job")
    return compile_job(job_settings, whitelisted_blocks)


def validate(job_settings: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None):
    """
    Validates a job in YAML 

    Args:
        job_settings (Dict[str, Any]): Job settings
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Raises:
        ValueError: When the job is invalid
    """
    logger.debug("Validating job")
    try:
        validate_job(
            source=job_settings,
            whitelisted_blocks=whitelisted_blocks
        )
    except Exception as e:
        raise ValueError(e)


def transform(job_settings: Dict[str, Any],
              data: List[Dict[str, Any]],
              context: Optional[Context] = None, whitelisted_blocks: Optional[List[str]] = None) -> List[Dict[str, Any]]:
    """
    Transforms data against a certain job

    Args:
        job_settings (Dict[str, Any]): Job settings
        data (List[Dict[str, Any]]): Data to transform
        context (Optional[Context], optional): Context. Defaults to None.
        whitelisted_blocks: (Optional[List[str]], optional): Whitelisted blocks. Defaults to None.

    Returns:
        List[Dict[str, Any]]: Transformed data
    """
    job = compile(job_settings, whitelisted_blocks)
    job.init(context)
    logger.debug("Transforming data")
    return job.transform(data)



import logging
from typing import Any, Dict, List, Tuple

from datayoga_core import utils
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Context = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.properties = utils.format_block_properties(self.properties)

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        logger.debug(f"Running {self.get_block_name()}")
        results = []

        for row in data:
            for property in self.properties["fields"]:
                key_found = True
                obj = row
                value = None
                from_field_path = utils.split_field(property["from_field"])

                for index, key in enumerate(from_field_path):
                    key = utils.unescape_field(key)

                    if key in obj:
                        if len(from_field_path) == index + 1:
                            value = obj[key]
                            del obj[key]
                        else:
                            obj = obj[key]
                    else:
                        key_found = False
                        break

                if key_found:
                    obj = row
                    to_field_path = utils.split_field(property["to_field"])

                    for key in to_field_path[:-1]:
                        key = utils.unescape_field(key)

                        if key in obj:
                            obj = obj[key]
                        else:
                            obj[key] = {}

                    if key in obj:
                        obj = obj[key]

                    obj[utils.unescape_field(to_field_path[-1:][0])] = value

        results.append(Result.SUCCESS)
        return data, results



import json
import os
import re
import sys
from os import path
from typing import Any, Dict, List

import yaml
from datayoga_core.context import Context


def read_json(filename: str) -> Any:
    """
    Loads a filename as a JSON object

    Args:
        filename (str): JSON filename to load

    Returns:
        Any: JSON object
    """
    with open(filename, "r", encoding="utf8") as f:
        return json.load(f)


def read_yaml(filename: str) -> Dict[str, Any]:
    """
    Loads a filename as a YAML object

    Args:
        filename (str): YAML filename to load

    Raises:
        ValueError: In case of invalid YAML

    Returns:
         Dict[str, Any]: YAML python object
    """
    try:
        with open(filename, "r", encoding="utf8") as stream:
            return yaml.safe_load(stream)
    except Exception as e:
        raise ValueError(f"Malformed YAML: {e}")


def format_block_properties(properties: Dict[str, Any]) -> Dict[str, Any]:
    """Adds `fields` array with the passed properties in case it's missing

    Args:
        properties (Dict[str, Any]): properties

    Returns:
        Dict[str, Any]: formatted properties with `fields` array
    """
    return {"fields": [properties]} if not "fields" in properties else properties


def is_bundled() -> bool:
    return getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")


def get_bundled_dir() -> str:
    datayoga_dir = path.join(sys._MEIPASS, "datayoga")
    return datayoga_dir if os.path.isdir(datayoga_dir) else sys._MEIPASS


def get_resource_path(relative_path: str) -> str:
    if is_bundled():
        # we are running in a bundle
        return path.join(get_bundled_dir(), "resources", relative_path)
    else:
        # we are running in a normal Python environment
        return path.join(os.path.dirname(__file__), "resources", relative_path)


def split_field(field: str) -> List[str]:
    return re.split(r"(?<!\\)\.", field)


def unescape_field(field: str) -> str:
    return field.replace("\\.", ".")


def get_connection_details(connection_name: str, context: Context) -> Dict[str, Any]:
    if context:
        connection = context.properties.get("connections").get(connection_name)
        if connection:
            return connection

    raise ValueError(f"{connection_name} connection not found")



import asyncio
import copy
import logging
import os
import sys
from enum import Enum
from typing import Any, Dict, List, Optional

import jsonschema
from datayoga_core import utils
from datayoga_core.block import Block, Result, create_block
from datayoga_core.context import Context
from datayoga_core.step import Step

logger = logging.getLogger("dy")


class ErrorHandling(Enum):
    ABORT = "abort"
    IGNORE = "ignore"


class Job():
    """
    Job

    Attributes:
        steps List[Block]: List of steps
    """

    def __init__(self, steps: Optional[List[Step]] = None, input: Optional[Block] = None,
                 error_handling: Optional[ErrorHandling] = None):
        """
        Constructs a job and its blocks

        Args:
            steps (List[Dict[str, Any]]): Job steps
            input (Optional[Block]): Block to be used as a producer
            error_handling (Optional[ErrorHandling]): error handling strategy
        """
        self.input = input
        self.steps = steps
        self.error_handling = error_handling if error_handling else ErrorHandling.IGNORE
        self.initialized = False

    def init(self, context: Optional[Context] = None):
        # open any connections or setup needed
        self.context = context
        self.root = None
        for step in self.steps:
            step.init(context)

            # create the step sequence
            if self.root is None:
                self.root = step
                last_step = self.root
            else:
                last_step = last_step.append(step)

        if self.input:
            self.input.init(context)

        self.root.add_done_callback(self.handle_result)
        self.initialized = True

    def transform(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """
        Transforms data

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        if not self.initialized:
            logger.debug("job has not been initialized yet, initializing...")
            self.init()

        transformed_data = copy.deepcopy(data)
        for step in self.steps:
            transformed_data, results = asyncio.run(step.block.run(transformed_data))
            logger.debug(transformed_data)

        return transformed_data

    async def run(self):
        for record in self.input.produce():
            logger.debug(f"Retrieved record:\n\t{record}")
            await self.root.process([record])

        await self.shutdown()

    async def shutdown(self):
        # wait for in-flight records to finish
        await self.root.join()

        # graceful shutdown
        await self.root.stop()

    def handle_result(self, msg_ids: List[str], result: Result, reason: str):
        if result == Result.REJECTED and self.error_handling == ErrorHandling.ABORT.value:
            logger.critical("Aborting due to rejected record(s)")
            sys.exit(1)

        self.input.ack(msg_ids)


#
# static utility methods
#

def validate_job(source: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None):
    # validate against the schema
    jsonschema.validate(instance=source, schema=utils.read_json(
        utils.get_resource_path(os.path.join("schemas", "job.schema.json"))))

    # validate that steps do not use any non whitelisted blocks
    if whitelisted_blocks is not None:
        for step in source.get("steps"):
            if step.get("uses") not in whitelisted_blocks:
                raise ValueError(f"use of invalid block type: {step.get('uses')}")


def compile_job(source: Dict[str, Any], whitelisted_blocks: Optional[List[str]] = None) -> Job:
    validate_job(source, whitelisted_blocks=whitelisted_blocks)
    steps: List[Step] = []
    # parse the steps
    for step_definition in source.get("steps"):
        block_type = step_definition.get("uses")
        block_properties = step_definition.get("with")
        block: Block = create_block(block_type, block_properties)
        step: Step = Step(step_definition.get("id", block_type), block)
        steps.append(step)

    # parse the input
    input = None
    if source.get("input") is not None:
        input_definition = source.get("input")
        input = create_block(input_definition.get("uses"), input_definition.get("with"))

    return Job(steps, input, source.get("error_handling"))



import json
import logging
from typing import Any, Dict, List, Tuple

from datayoga_core import expression
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Context = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.expression = expression.compile(self.properties["language"], json.dumps(self.properties["expression"]))

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        logger.debug(f"Running {self.get_block_name()}")
        results = []
        return_data = []
        for row in data:
            mapped_row = self.expression.search(row)
            # we always add the msg_id back
            if Block.MSG_ID_FIELD in row:
                mapped_row[Block.MSG_ID_FIELD] = row[Block.MSG_ID_FIELD]
            return_data.append(mapped_row)

        results.append(Result.SUCCESS)
        return return_data, results



import logging
from typing import Any, Dict, Generator, List, TypedDict

from .block import Block

logger = logging.getLogger("dy")
Message = TypedDict("Message", msg_id=str, value=Dict[str, Any])


class Producer(Block):

    def produce(self) -> Generator[Message, None, None]:
        """ Produces data (abstract, should be implemented by the sub class)

        Returns:
            Generator[Message]: Produced data
        """
        pass

    def ack(self, msg_ids: List[str]):
        """ Sends acknowledge for the message IDs of the records that have been processed (abstract, should be implemented by the sub class)

        Args:
            msg_ids (List[str]): Message IDs
        """
        pass



import logging
from typing import Any, Dict, List, Tuple

from datayoga_core import expression
from datayoga_core.block import Block as DyBlock
from datayoga_core.block import Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Block(DyBlock):
    def init(self, context: Context = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        self.expression = expression.compile(self.properties["language"], self.properties["expression"])

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        logger.debug(f"Running {self.get_block_name()}")
        return_data = self.expression.filter(data)
        return return_data, [Result.SUCCESS] * len(return_data)



import hashlib
import string
from datetime import datetime, timezone
from typing import Union
from uuid import uuid4

import json
from jmespath import functions


# custom functions for Jmespath
class JmespathCustomFunctions(functions.Functions):

    @functions.signature({"types": ["string", "null"]})
    def _func_capitalize(self, arg):
        return string.capwords(str(arg)) if arg is not None else None

    @functions.signature({"types": ["array"]})
    def _func_concat(self, elements):
        return ''.join([str(x) for x in elements])

    @functions.signature({"types": ["string", "null"]})
    def _func_lower(self, element):
        return str(element).lower() if element is not None else None

    @functions.signature({"types": ["string", "null"]})
    def _func_upper(self, element):
        return str(element).upper() if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["string"]}, {"types": ["string"]})
    def _func_replace(self, element, old_value, new_value):
        return str(element).replace(old_value, new_value) if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]})
    def _func_right(self, element, amount):
        return str(element)[-amount:] if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]})
    def _func_left(self, element, amount):
        return str(element)[:amount] if element is not None else None

    @functions.signature({"types": ["string", "null"]}, {"types": ["number"]}, {"types": ["number"]})
    def _func_mid(self, element, offset, amount):
        return str(element)[offset:offset+amount] if element is not None else None

    @functions.signature({"types": ["string", "null"], "variadic": True})
    def _func_split(self, element, delimiter=","):
        return str(element).split(delimiter) if element is not None else None

    @functions.signature()
    def _func_uuid(self):
        """Generates a random UUID4 and returns it as a string in standard format."""

        return str(uuid4())

    @functions.signature({"types": ["number", "string", "boolean", "array", "object", "null"], "variadic": True})
    def _func_hash(self, obj, hash_name="sha1"):
        """\
        Calculates a hash using given the `hash_name` hash function and returns its hexadecimal representation.

        Supported algorithms:

        - sha1(default)
        - sha256
        - md5
        - sha384
        - sha3_384
        - blake2b
        - sha512
        - sha3_224
        - sha224
        - sha3_256
        - sha3_512
        - blake2s

        See https://docs.python.org/3/library/hashlib.html for more information.
        """

        def prepare() -> Union[bytes, bytearray]:
            if isinstance(obj, (bytes, bytearray)):
                return obj

            if obj is None:
                return b""

            if isinstance(obj, str):
                return obj.encode()

            # the 'separators' arg is needed to remove whitespace in the resulting string
            return json.dumps(obj, separators=(',', ':')).encode("utf-8")

        h = hashlib.new(hash_name)
        h.update(prepare())
        return h.hexdigest()

    @functions.signature({"types": ["string", "number"]})
    def _func_time_delta_days(self, dt):
        """\
        Returns the number of days between given `dt` and now (positive)
        or the number of days that have passed from now (negative).

        If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed.
        If `dt` is a number, Unix timestamp (1320365123, for example) is assumed.
        """

        dt = datetime.fromisoformat(dt) if isinstance(dt, str) else datetime.fromtimestamp(dt, timezone.utc)
        delta = dt.now(dt.tzinfo) - dt
        return delta.days

    @functions.signature({"types": ["string", "number"]})
    def _func_time_delta_seconds(self, dt):
        """\
        Returns the number of seconds between given `dt` and now (positive)
        or the number of seconds that have passed from now (negative).

        If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed.
        If `dt` is a number, Unix timestamp (1320365123, for example) is assumed.
        """

        dt = datetime.fromisoformat(dt) if isinstance(dt, str) else datetime.fromtimestamp(dt, timezone.utc)
        delta = dt.now(dt.tzinfo) - dt

        return delta.days * 86400 + delta.seconds



import pytest
from datayoga_core.block import Result
from datayoga_core.blocks.remove_field.block import Block


@pytest.mark.asyncio
async def test_remove_existing_field():
    block = Block({"field": "fname"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == ([{"lname": "doe"}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_missing_field():
    block = Block({"field": "mname"})
    block.init()
    assert await block.run([{"fname": "john", "lname": "doe"}]) == ([{"fname": "john", "lname": "doe"}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_deep_nested_field():
    block = Block({"field": "employee.name.fname"})
    block.init()
    assert await block.run([{"employee": {"name": {"fname": "john", "lname": "doe"}}}]) == ([
        {"employee": {"name": {"lname": "doe"}}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_nested_field():
    block = Block({"field": "name.fname"})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == ([{"name": {"lname": "doe"}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_missing_nested_field():
    block = Block({"field": "full_name.fname"})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == ([{"name": {"fname": "john", "lname": "doe"}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_multiple_fields():
    block = Block({"fields": [{"field": "name.fname"}, {"field": "name.lname"}]})
    block.init()
    assert await block.run([{"name": {"fname": "john", "lname": "doe"}}]) == ([{"name": {}}], [Result.SUCCESS])


@pytest.mark.asyncio
async def test_remove_field_with_dot():
    block = Block({"field": "name\.fname"})
    block.init()
    assert await block.run([{"name.fname": "john", "lname": "doe"}]) == ([{"lname": "doe"}], [Result.SUCCESS])



from __future__ import annotations

import asyncio
import logging
from typing import Any, Callable, Dict, List, Optional

from datayoga_core.block import Block, Result
from datayoga_core.context import Context

logger = logging.getLogger("dy")


class Step():
    def __init__(self, id: str, block: Block, concurrency=1):
        self.id = id
        self.block = block
        self.next_step = None
        self.active_entries = set()
        self.concurrency = concurrency
        self.workers = [None]*self.concurrency
        self.done_callback = None
        self.initialized = False

    def init(self, context: Optional[Context] = None):
        # initialize the block
        self.block.init(context)
        # start pool of workers for parallelization

    async def start_pool(self):
        logger.debug("starting pool")
        self.queue = asyncio.Queue(maxsize=1)
        for id in range(self.concurrency):
            worker = self.workers[id]
            if worker is None or not worker.done():
                self.workers[id] = asyncio.create_task(self.run(id))
            else:
                logger.debug(f"worker {id} is running: {not worker.done()}")

    def add_done_callback(self, callback: Callable[[str], None]):
        self.done_callback = callback

    def __or__(self, other: Step):
        return self.append(other)

    def append(self, next_step: Step):
        self.next_step = next_step
        self.next_step.add_done_callback(self.done)
        return self.next_step

    async def process(self, messages: List[Dict[str, Any]]):
        if not self.initialized:
            await self.start_pool()
            self.initialized = True

        self.active_entries.update([x[Block.MSG_ID_FIELD] for x in messages])
        await self.queue.put(messages)

    async def run(self, worker_id: int):
        while True:
            entry = await self.queue.get()
            logger.debug(f"{self.id}-{worker_id} processing {[i[Block.MSG_ID_FIELD] for i in entry]}")
            try:
                processed_entries, results = await self.block.run(entry)

                # TODO: handle filtered. anything not processed or rejected
                # check if we have a next step
                if self.next_step:
                    # process downstream
                    await self.next_step.process(processed_entries)
                else:
                    # we are a last channel, propagate the ack upstream
                    # TODO: verify that all entries have a msg id otherwise raise consistency error
                    self.done([x[Block.MSG_ID_FIELD] for x in processed_entries], Result.SUCCESS)

                    # indicate rejected records if any
                    if Result.REJECTED in results:
                        self.done([entry[i][Block.MSG_ID_FIELD]
                                  for i, v in enumerate(results) if v == Result.REJECTED], Result.REJECTED)
            except Exception as e:
                # we caught an exception. the entire batch is considered rejected
                logger.exception(e)
                # verify that all messages still have a msg_id property
                # if next(filter(lambda x: not Block.MSG_ID_FIELD in x,entry),None) is not None
                self.done([x[Block.MSG_ID_FIELD] for x in entry],
                          Result.REJECTED, f"Error in step {self.id}: {repr(e)}")
            finally:
                self.queue.task_done()
            logger.debug(f"{self.id}-{worker_id} done processing {entry[0][Block.MSG_ID_FIELD]}")

    def done(self, msg_ids: List[str], result: Optional[Result] = None, reason: Optional[str] = None):
        logger.debug(f"{self.id} acking {msg_ids} with result {result}")
        self.active_entries.difference_update(msg_ids)
        if self.done_callback is not None:
            self.done_callback(msg_ids, result, reason)

    async def join(self):
        # wait for all active entries to be processed
        while len(self.active_entries) > 0:
            await asyncio.sleep(0.2)

    async def stop(self):
        # wait for all tasks to finish
        await self.join()

        # stop any downstream workers
        if self.next_step:
            await self.next_step.stop()

        # stop all workers in the pool
        for worker in self.workers:
            if worker:
                worker.cancel()



import logging
import os
from csv import DictReader
from typing import Generator, Optional

from datayoga_core.context import Context
from datayoga_core.producer import Message
from datayoga_core.producer import Producer as DyProducer

logger = logging.getLogger("dy")


class Block(DyProducer):

    def init(self, context: Optional[Context] = None):
        logger.debug(f"Initializing {self.get_block_name()}")
        csv_file = self.properties["file"]

        if os.path.isabs(csv_file) or context is None:
            self.file = csv_file
        else:
            self.file = os.path.join(context.properties.get("data_path"), csv_file)

        logger.debug(f"file: {self.file}")
        self.batch_size = self.properties.get("batch_size", 1000)

    def produce(self) -> Generator[Message, None, None]:
        logger.debug("Reading CSV")

        with open(self.file, "r") as read_obj:
            records = list(DictReader(read_obj))

        for i, record in enumerate(records):
            yield {self.MSG_ID_FIELD: str(i), **record}



import importlib
import logging
import os
import sys
from enum import Enum
from os import path
from typing import Any, Dict, List, Optional, Tuple

from datayoga_core import utils
from datayoga_core.context import Context
from jsonschema import validate

logger = logging.getLogger("dy")


Result = Enum("Result", "SUCCESS REJECTED FILTERED")


class Block():
    MSG_ID_FIELD = "__$$msg_id"
    """
    Block

    Attributes:
        properties Dict[str, Any]: Block properties
    """

    def __init__(self, properties: Dict[str, Any] = {}):
        """
        Constructs a block

        Args:
            properties (Dict[str, Any]): Block [properties]
        """
        self.properties = properties
        self.validate()

    def validate(self):
        """
        Validates block against its JSON Schema
        """
        json_schema_file = path.join(
            utils.get_bundled_dir(), "blocks", self.get_block_name(),
            "block.schema.json") if utils.is_bundled() else path.join(
            os.path.dirname(os.path.realpath(sys.modules[self.__module__].__file__)),
            "block.schema.json")

        logger.debug(f"validating {self.properties} against {json_schema_file}")
        validate(instance=self.properties, schema=utils.read_json(json_schema_file))

    def init(self, context: Optional[Context] = None):
        """
        Initializes block (abstract, should be implemented by the sub class)

        Args:
            context (Context, optional): Context. Defaults to None.
        """
        pass

    async def run(self, data: List[Dict[str, Any]]) -> Tuple[List[Dict[str, Any]], List[Result]]:
        """ Transforms data (abstract, should be implemented by the sub class)

        Args:
            data (List[Dict[str, Any]]): Data

        Returns:
            List[Dict[str, Any]]: Transformed data
        """
        pass

    def get_block_name(self):
        return os.path.basename(os.path.dirname(sys.modules[self.__module__].__file__))


#
# static utility methods
#

def create_block(block_name: str, properties: Dict[str, Any]) -> Block:
    module_name = f"datayoga_core.blocks.{block_name}.block"
    module = importlib.import_module(module_name)
    block: Block = getattr(module, "Block")(properties)
    return block



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['datayoga_core',
 'datayoga_core.blocks',
 'datayoga_core.blocks.add_field',
 'datayoga_core.blocks.add_field.tests',
 'datayoga_core.blocks.files.read_csv',
 'datayoga_core.blocks.filter',
 'datayoga_core.blocks.filter.tests',
 'datayoga_core.blocks.map',
 'datayoga_core.blocks.redis',
 'datayoga_core.blocks.redis.read_stream',
 'datayoga_core.blocks.redis.write',
 'datayoga_core.blocks.relational.write',
 'datayoga_core.blocks.remove_field',
 'datayoga_core.blocks.remove_field.tests',
 'datayoga_core.blocks.rename_field',
 'datayoga_core.blocks.rename_field.tests',
 'datayoga_core.blocks.sequence',
 'datayoga_core.blocks.std.read',
 'datayoga_core.blocks.std.write']

package_data = \
{'': ['*'],
 'datayoga_core': ['resources/scaffold/*',
                   'resources/scaffold/data/*',
                   'resources/scaffold/jobs/sample/*',
                   'resources/schemas/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0.3,<9.0.0',
 'jmespath>=1.0.0,<2.0.0',
 'jsonschema>=4.4.0,<5.0.0']

extras_require = \
{'pg': ['psycopg2>=2.9.3,<3.0.0', 'SQLAlchemy>=1.4.41,<2.0.0'],
 'redis': ['redis>=4.1.4,<5.0.0'],
 'test': ['mock>=4.0.3,<5.0.0',
          'pytest>=7.1.2,<8.0.0',
          'pytest-describe>=2.0.1,<3.0.0',
          'pytest-mock>=3.7.0,<4.0.0',
          'pytest-timeout>=2.1.0,<3.0.0',
          'requests-mock>=1.9.3,<2.0.0',
          'testcontainers>=3.7.0,<4.0.0',
          'psycopg2>=2.9.3,<3.0.0',
          'redis>=4.1.4,<5.0.0',
          'SQLAlchemy>=1.4.41,<2.0.0']}

setup_kwargs = {
    'name': 'datayoga-core',
    'version': '1.9.0',
    'description': 'DataYoga for Python',
    'long_description': '# DataYoga Core\n\n## Introduction\n\n`datayoga-core` is the transformation engine used in `DataYoga`, a framework for building and generating data pipelines.\n\n## Installation\n\n```bash\npip install datayoga-core\n```\n\n## Quick Start\n\nThis demonstrates how to transform data using a DataYoga job.\n\n### Create a Job\n\nUse this `example.yaml`:\n\n```yaml\n- steps:\n    - uses: add_field\n      with:\n        fields:\n          - field: full_name\n            language: jmespath\n            expression: concat([fname, \' \' , lname])\n          - field: country\n            language: sql\n            expression: country_code || \' - \' || UPPER(country_name)\n    - uses: rename_field\n      with:\n        fields:\n          - from_field: fname\n            to_field: first_name\n          - from_field: lname\n            to_field: last_name\n    - uses: remove_field\n      with:\n        fields:\n          - field: credit_card\n          - field: country_name\n          - field: country_code\n    - uses: map\n      with:\n        expression:\n          {\n            first_name: first_name,\n            last_name: last_name,\n            greeting: "\'Hello \' || CASE WHEN gender = \'F\' THEN \'Ms.\' WHEN gender = \'M\' THEN \'Mr.\' ELSE \'N/A\' END || \' \' || full_name",\n            country: country,\n            full_name: full_name,\n          }\n        language: sql\n```\n\n### Transform Data Using `datayoga-core`\n\nUse this code snippet to transform a data record using the job defined [above](#create-a-job):\n\n```python\nimport datayoga_core as dy\nfrom datayoga_core.job import Job\nfrom datayoga_core.utils import read_yaml\n\njob_settings = read_yaml("example.yaml")\njob = dy.compile(job_settings)\n\nassert job.transform({"fname": "jane", "lname": "smith", "country_code": 1, "country_name": "usa", "credit_card": "1234-5678-0000-9999", "gender": "F"}) == {"first_name": "jane", "last_name": "smith", "country": "1 - USA", "full_name": "jane smith", "greeting": "Hello Ms. jane smith"}\n```\n\nAs can be seen, the record has been transformed based on the job:\n\n- `fname` field renamed to `first_name`.\n- `lname` field renamed to `last_name`.\n- `country` field added based on an SQL expression.\n- `full_name` field added based on a [JMESPath](https://jmespath.org/) expression.\n- `greeting` field added based on an SQL expression.\n\n### Examples\n\n- Add a new field `country` out of an SQL expression that concatenates `country_code` and `country_name` fields after upper case the later:\n\n  ```yaml\n  uses: add_field\n  with:\n    field: country\n    language: sql\n    expression: country_code || \' - \' || UPPER(country_name)\n  ```\n\n- Rename `fname` field to `first_name` and `lname` field to `last_name`:\n\n  ```yaml\n  uses: rename_field\n  with:\n    fields:\n      - from_field: fname\n        to_field: first_name\n      - from_field: lname\n        to_field: last_name\n  ```\n\n- Remove `credit_card` field:\n\n  ```yaml\n  uses: remove_field\n  with:\n    field: credit_card\n  ```\n\nFor a full list of supported block types [see reference](https://datayoga-io.github.io/datayoga-py/).\n\n## Expression Language\n\nDataYoga supports both SQL and [JMESPath](https://jmespath.org/) expressions. JMESPath are especially useful to handle nested JSON data, while SQL is more suited to flat row-like structures.\n\n### Notes\n\n- Dot notation in expression represents nesting fields in the object, for example `name.first_name` refers to `{ "name": { "first_name": "John" } }`.\n- In order to refer to a field that contains a dot in its name, escape it, for example `name\\.first_name` refers to `{ "name.first_name": "John" }`.\n\n### JMESPath Custom Functions\n\nDataYoga adds the following custom functions to the standard JMESPath library:\n\n| Function             | Description                                                                                                                         | Example                                                                                                                                           | Comments                                                                                                                                                         |\n| -------------------- | ----------------------------------------------------------------------------------------------------------------------------------- | ------------------------------------------------------------------------------------------------------------------------------------------------- | ---------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| `capitalize`         | Capitalizes all the words in the string                                                                                             | Input: `{"name": "john doe"}` <br /> Expression: `capitalize(name)` <br /> Output: `John Doe`                                                     |                                                                                                                                                                  |\n| `concat`             | Concatenates an array of variables or literals                                                                                      | Input: `{"fname": "john", "lname": "doe"}` <br /> Expression: `concat([fname, \' \' ,lname])` <br /> Output: `john doe`                             | This is equivalent to the more verbose built-in expression: `\' \'.join([fname,lname])`                                                                            |\n| `hash`               | Calculates a hash using the `hash_name` hash function and returns its hexadecimal representation                                    | Input: `{"some_str": "some_value"}` <br /> Expression: `hash(some_str, &#96;sha1&#96;)` <br /> Output: `8c818171573b03feeae08b0b4ffeb6999e3afc05` | Supported algorithms: sha1 (default), sha256, md5, sha384, sha3_384, blake2b, sha512, sha3_224, sha224, sha3_256, sha3_512, blake2s                              |\n| `left`               | Returns a specified number of characters from the start of a given text string                                                      | Input: `{"greeting": "hello world!"}` <br /> Expression: `` left(greeting, `5`) `` <br /> Output: `hello`                                         |                                                                                                                                                                  |\n| `lower`              | Converts all uppercase characters in a string into lowercase characters                                                             | Input: `{"fname": "John"}` <br /> Expression: `lower(fname)` <br /> Output: `john`                                                                |                                                                                                                                                                  |\n| `mid `               | Returns a specified number of characters from the middle of a given text string                                                     | Input: `{"greeting": "hello world!"}` <br /> Expression: `` mid(greeting, `4`, `3`) `` <br /> Output: `o w`                                       |                                                                                                                                                                  |\n| `replace`            | Replaces all the occurrences of a substring with a new one                                                                          | Input: `{"sentence": "one four three four!"}` <br /> Expression: `replace(sentence, \'four\', \'two\')` <br /> Output: `one two three two!`           |                                                                                                                                                                  |\n| `right`              | Returns a specified number of characters from the end of a given text string                                                        | Input: `{"greeting": "hello world!"}` <br /> Expression: `` right(greeting, `6`) `` <br /> Output: `world!`                                       |                                                                                                                                                                  |\n| `split`              | Splits a string into a list of strings after breaking the given string by the specified delimiter (comma by default)                | Input: `{"departments": "finance,hr,r&d"}` <br /> Expression: `split(departments)` <br /> Output: `[\'finance\', \'hr\', \'r&d\']`                      | Default delimiter is comma - a different delimiter can be passed to the function as the second argument, for example: `split(departments, \';\')`                  |\n| `time_delta_days`    | Returns the number of days between a given `dt` and now (positive) or the number of days that have passed from now (negative)       | Input: `{"dt": \'2021-10-06T18:56:16.701670+00:00\'}` <br /> Expression: `time_delta_days(dt)` <br /> Output: `365`                                 | If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed. If `dt` is a number, Unix timestamp (1320365123, for example) is assumed. |\n| `time_delta_seconds` | Returns the number of seconds between a given `dt` and now (positive) or the number of seconds that have passed from now (negative) | Input: `{"dt": \'2021-10-06T18:56:16.701670+00:00\'}` <br /> Expression: `time_delta_days(dt)` <br /> Output: `31557600`                            | If `dt` is a string, ISO datetime (2011-11-04T00:05:23+04:00, for example) is assumed. If `dt` is a number, Unix timestamp (1320365123, for example) is assumed. |\n| `upper`              | Converts all lowercase characters in a string into uppercase characters                                                             | Input: `{"fname": "john"}` <br /> Expression: `upper(fname)` <br /> Output: `JOHN`                                                                |                                                                                                                                                                  |\n| `uuid`               | Generates a random UUID4 and returns it as a string in standard format                                                              | Input: None <br /> Expression: `uuid()` <br /> Output: `3264b35c-ff5d-44a8-8bc7-9be409dac2b7`                                                     |                                                                                                                                                                  |\n',
    'author': 'DataYoga',
    'author_email': 'admin@datayoga.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, Optional


class Context():
    """
    Context

    Attributes:
        properties Dict[str, Any]: Context properties
        state (Dict[str, Any]): Context state
    """

    def __init__(self, properties: Optional[Dict[str, Any]] = None, state: Optional[Dict[str, Any]] = None):
        """
        Constructs a context

        Args:
            properties (Optional[Dict[str, Any]], optional): Context properties. Defaults to None.
            state (Optional[Dict[str, Any]], optional): Context state. Defaults to None.
        """
        self.properties = properties
        self.state = state
